#include "fastmath.h"
#include <cstdlib>

#include "chunk.h"

Chunk::Chunk(int x, int y, int z, World *parent)
    : x(x), y(y), z(z), abs_x(x*SIZE*BLOCK_SIZE), abs_y(y*SIZE*BLOCK_SIZE), abs_z(z*SIZE*BLOCK_SIZE), parent(parent), aabb(abs_x, abs_y, abs_z, abs_x + SIZE*BLOCK_SIZE, abs_y + SIZE*BLOCK_SIZE, abs_z + SIZE*BLOCK_SIZE)
{
    //Everything air
    memset(blocks, BLOCK_AIR, sizeof(blocks));

    printf("Generating chunk %d:%d:%d...\t", x, y, z);

    for(int x = 0; x < SIZE; x++)
        for(int z = 0; z < SIZE; z++)
        {
            int height = (fast_sin(GLFix(x*180)/SIZE) * fast_sin(GLFix(z*180)/SIZE) + 1) * 4 + 1;
            for(int y = 0; y < height; y++)
                blocks[x][y][z] = y < 2 ? BLOCK_STONE : (height > 6 ? BLOCK_DIRT : BLOCK_SAND);

            if(x == this->x + SIZE/2 && z == this->z + SIZE/2)
                makeTree(x, height, z);
        }

    printf("Done!\n");
}

RGB operator *(const RGB& rgb, const GLFix f)
{
    return { rgb.r * f, rgb.g * f, rgb.b * f };
}

COLOR operator *(const COLOR c, const GLFix f)
{
    RGB rgb = rgbColor(c) * f;
    return colorRGB(rgb.r, rgb.g, rgb.b);
}

#define IN_BOUNDS(x, y, z) (x >= 0 && y >= 0 && z >= 0 && x < SIZE && y < SIZE && z < SIZE)

void Chunk::generateLight(int x, int y, int z, uint8_t value)
{
    if(x < -1 || x > SIZE || y < -1 || y > SIZE || z < -1 || z > SIZE)
        return;

    if(IN_BOUNDS(x, y, z) && blocks[x][y][z] != BLOCK_AIR)
        return;

    if(getLight(x, y, z) >= value)
        return;

    setLight(x, y, z, value);

    uint8_t new_light = value - 5;
    generateLight(x - 1, y, z, new_light);
    generateLight(x + 1, y, z, new_light);
    generateLight(x, y - 1, z, new_light);
    generateLight(x, y + 1, z, new_light);
    generateLight(x, y, z - 1, new_light);
    generateLight(x, y, z + 1, new_light);
}

int Chunk::getPosition(GLFix x, GLFix y, GLFix z)
{
    if(pos_indices[x][y][z] == -1)
    {
        pos_indices[x][y][z] = positions.size();
        positions.emplace_back(Position{abs_x + x*BLOCK_SIZE, abs_y + y*BLOCK_SIZE, abs_z + z*BLOCK_SIZE});
    }

    return pos_indices[x][y][z];
}

void Chunk::addVertex(const GLFix x, const GLFix y, const GLFix z, GLFix u, GLFix v, const COLOR c)
{
#ifdef TEXTURE_SUPPORT
    //We skip the texture transformations nglTransformVertex does,
    //So we have to do them here

    const TEXTURE *texture = nglGetTexture();

    u *= texture->width;
    v *= texture->height;

    if(u >= GLFix(texture->width))
        u = texture->width - 1;
    if(v == GLFix(texture->height))
        v = texture->height - 1;
#endif

    vertices.emplace_back(IndexedVertex{getPosition(x, y, z), u, v, c});
}

void Chunk::buildGeometry()
{
    __builtin_memset(pos_indices, -1, sizeof(pos_indices));

    positions.clear();
    vertices.clear();

    printf("Updating chunk %d:%d:%d...\n", x, y, z);

    #ifdef LIGHTING
        printf("\tGenerating light...\t");
        memset(light, 0, sizeof(light));

        for(int x = 0; x < SIZE; x++)
            for(int z = 0; z < SIZE; z++)
                for(int y = SIZE; y > 1; y--)
                {
                    if(blocks[x][y - 1][z] == BLOCK_AIR)
                        continue;

                    generateLight(x, y, z, 30);
                    break;
                }
        printf("Done!\n");
    #endif

    printf("\tUpdating geometry...\t");
    int count_indices = 0;

    //Calculate size of buffer
    for(int x = -1; x <= SIZE; x++)
        for(int y = -1; y <= SIZE; y++)
            for(int z = -1; z <= SIZE; z++)
            {
                //If not air, sides of adjacent blocks aren't visible
                if(getGlobalBlockRelative(x, y, z) != BLOCK_AIR)
                    continue;

                if(IN_BOUNDS(x - 1, y, z) && blocks[x - 1][y][z] != BLOCK_AIR)
                    count_indices += 4;

                if(IN_BOUNDS(x + 1, y, z) && blocks[x + 1][y][z] != BLOCK_AIR)
                    count_indices += 4;

                if(IN_BOUNDS(x, y - 1, z) && blocks[x][y - 1][z] != BLOCK_AIR)
                    count_indices += 4;

                if(IN_BOUNDS(x, y + 1, z) && blocks[x][y + 1][z] != BLOCK_AIR)
                    count_indices += 4;

                if(IN_BOUNDS(x, y, z - 1) && blocks[x][y][z - 1] != BLOCK_AIR)
                    count_indices += 4;

                if(IN_BOUNDS(x, y, z + 1) && blocks[x][y][z + 1] != BLOCK_AIR)
                    count_indices += 4;
            }

    vertices.reserve(count_indices);

    //Now, render!
    GLFix posX = -1;

    for(int x = -1; x <= SIZE; x++, posX += 1)
    {
        GLFix posY = -1;
        for(int y = -1; y <= SIZE; y++, posY += 1)
        {
            GLFix posZ = -1;
            for(int z = -1; z <= SIZE; z++, posZ += 1)
            {
                //If not air, sides of adjacent blocks aren't visible
                if(getGlobalBlockRelative(x, y, z) != BLOCK_AIR)
                    continue;

                GLFix light_level = 0xFFFF;
                #ifdef LIGHTING
                    light_level = lightToFix(getLight(x, y, z));
                #endif

                if((x + z) % 2 == 0)
                    light_level *= 0.8f;

                if(IN_BOUNDS(x - 1, y, z) && blocks[x - 1][y][z] != BLOCK_AIR)
                {
                    GLFix myposX = posX - 1;
                    GLFix myposY = posY;
                    GLFix myposZ = posZ;

                    #ifndef TEXTURE_SUPPORT
                        COLOR c = block_colors[blocks[x - 1][y][z]];
                        c = c * light_level;
                    #else
                        COLOR c = 0;
                    #endif

                    TextureAtlasEntry &ri = block_textures[blocks[x - 1][y][z]];
                    addVertex(myposX + 1, myposY, myposZ, ri.right, ri.bottom, c);
                    addVertex(myposX + 1, myposY + 1, myposZ, ri.right, ri.top, c);
                    addVertex(myposX + 1, myposY + 1, myposZ + 1, ri.left, ri.top, c);
                    addVertex(myposX + 1, myposY, myposZ + 1, ri.left, ri.bottom, c);
                }

                if(IN_BOUNDS(x + 1, y, z) && blocks[x + 1][y][z] != BLOCK_AIR)
                {
                    GLFix myposX = posX + 1;
                    GLFix myposY = posY;
                    GLFix myposZ = posZ;

                    #ifndef TEXTURE_SUPPORT
                        COLOR c = block_colors[blocks[x + 1][y][z]];
                        c = c * light_level;
                    #else
                        COLOR c = 0;
                    #endif

                    TextureAtlasEntry &le = block_textures[blocks[x + 1][y][z]];
                    addVertex(myposX, myposY, myposZ + 1, le.left, le.bottom, c);
                    addVertex(myposX, myposY + 1, myposZ + 1, le.left, le.top, c);
                    addVertex(myposX, myposY + 1, myposZ, le.right, le.top, c);
                    addVertex(myposX, myposY, myposZ, le.right, le.bottom, c);
                }

                if(IN_BOUNDS(x, y - 1, z) && blocks[x][y - 1][z] != BLOCK_AIR)
                {
                    GLFix myposX = posX;
                    GLFix myposY = posY - 1;
                    GLFix myposZ = posZ;

                    #ifndef TEXTURE_SUPPORT
                        COLOR c = block_colors[blocks[x][y - 1][z]];
                        c = c * light_level;
                    #else
                        COLOR c = 0;
                    #endif

                    TextureAtlasEntry &to = block_textures[blocks[x][y - 1][z]];
                    addVertex(myposX, myposY + 1, myposZ, to.left, to.bottom, c);
                    addVertex(myposX, myposY + 1, myposZ + 1, to.left, to.top, c);
                    addVertex(myposX + 1, myposY + 1, myposZ + 1, to.right, to.top, c);
                    addVertex(myposX + 1, myposY + 1, myposZ, to.right, to.bottom, c);
                }

                if(IN_BOUNDS(x, y + 1, z) && blocks[x][y + 1][z] != BLOCK_AIR)
                {
                    GLFix myposX = posX;
                    GLFix myposY = posY + 1;
                    GLFix myposZ = posZ;

                    #ifndef TEXTURE_SUPPORT
                        COLOR c = block_colors[blocks[x][y + 1][z]];
                        c = c * light_level;
                    #else
                        COLOR c = 0;
                    #endif

                    #ifdef LIGHTING
                        c = c * light_level;
                    #endif

                    TextureAtlasEntry &bo = block_textures[blocks[x][y + 1][z]];
                    addVertex(myposX + 1, myposY, myposZ, bo.left, bo.bottom, c);
                    addVertex(myposX + 1, myposY, myposZ + 1, bo.left, bo.top, c);
                    addVertex(myposX, myposY, myposZ + 1, bo.right, bo.top, c);
                    addVertex(myposX, myposY, myposZ, bo.right, bo.bottom, c);
                }

                if(IN_BOUNDS(x, y, z - 1) && blocks[x][y][z - 1] != BLOCK_AIR)
                {
                    GLFix myposX = posX;
                    GLFix myposY = posY;
                    GLFix myposZ = posZ - 1;

                    #ifndef TEXTURE_SUPPORT
                        COLOR c = block_colors[blocks[x][y][z - 1]];
                        c = c * light_level;
                    #else
                        COLOR c = 0;
                    #endif

                    TextureAtlasEntry &ba = block_textures[blocks[x][y][z - 1]];
                    addVertex(myposX + 1, myposY, myposZ + 1, ba.left, ba.bottom, c);
                    addVertex(myposX + 1, myposY + 1, myposZ + 1, ba.left, ba.top, c);
                    addVertex(myposX, myposY + 1, myposZ + 1, ba.right, ba.top, c);
                    addVertex(myposX, myposY, myposZ + 1, ba.right, ba.bottom, c);
                }

                if(IN_BOUNDS(x, y, z + 1) && blocks[x][y][z + 1] != BLOCK_AIR)
                {
                    GLFix myposX = posX;
                    GLFix myposY = posY;
                    GLFix myposZ = posZ + 1;

                    #ifndef TEXTURE_SUPPORT
                        COLOR c = block_colors[blocks[x][y][z + 1]];
                        c = c * light_level;
                    #else
                        COLOR c = 0;
                    #endif

                    TextureAtlasEntry &fr = block_textures[blocks[x][y][z + 1]];
                    addVertex(myposX, myposY, myposZ, fr.left, fr.bottom, c);
                    addVertex(myposX, myposY + 1, myposZ, fr.left, fr.top, c);
                    addVertex(myposX + 1, myposY + 1, myposZ, fr.right, fr.top, c);
                    addVertex(myposX + 1, myposY, myposZ, fr.right, fr.bottom, c);
                }
            }
        }

        positions_transformed.resize(positions.size());
    }

    render_dirty = false;

    printf("Done!\n");
}

void Chunk::render()
{
    if(__builtin_expect(render_dirty, 0))
        buildGeometry();

    for(unsigned int i = 0; i < positions.size(); i++)
    {
        Position &position = positions.at(i);
        VERTEX t1{position.x, position.y, position.z, 0, 0, 0}, t2;
        nglTransformVertex(&t1, &t2);
        positions_transformed.at(i) = {t2.x, t2.y, t2.z};
    }

    IndexedVertex *v = vertices.data();
    for(unsigned int i = 0; i < vertices.size(); i += 4, v += 4)
    {
        VERTEX quad[4];

        for(unsigned int j = 0; j < 3; j++)
        {
            IndexedVertex &v1 = v[j];
            Position &position = positions_transformed.at(v1.pos);
            quad[j] = {position.x, position.y, position.z, v1.u, v1.v, v1.c};
        }

        GLFix x1 = quad[1].x - quad[0].x, x2 = quad[2].x - quad[0].x;
        GLFix y1 = quad[1].y - quad[0].y, y2 = quad[2].y - quad[0].y;

        if(y1 * x2 >= y2 * x1)
            continue;

        Position &position = positions_transformed.at(v[3].pos);
        quad[3] = {position.x, position.y, position.z, v[3].u, v[3].v, v[3].c};

        #ifdef WIREFRAME_MODE
                nglDrawLine3D(&quad[0], &quad[1]);
                nglDrawLine3D(&quad[1], &quad[2]);
                nglDrawLine3D(&quad[2], &quad[3]);
                nglDrawLine3D(&quad[3], &quad[0]);
        #else
                nglDrawTriangle(&quad[0], &quad[1], &quad[2]);
                nglDrawTriangle(&quad[2], &quad[3], &quad[0]);
        #endif
    }
}

BLOCK Chunk::getLocalBlock(int x, int y, int z)
{
    return blocks[x][y][z];
}

BLOCK Chunk::getGlobalBlockRelative(int x, int y, int z)
{
    if(IN_BOUNDS(x, y, z))
        return getLocalBlock(x, y, z);

    return parent->getBlock(x + this->x*SIZE, y + this->y*SIZE, z + this->z*SIZE);
}

bool Chunk::intersects(AABB &other)
{
    if(!aabb.intersects(other))
        return false;

    AABB aabb;
    aabb.low_x = this->x*SIZE*BLOCK_SIZE;

    for(unsigned int x = 0; x < SIZE; x++)
    {
        aabb.high_x = aabb.low_x + BLOCK_SIZE;

        aabb.low_y = this->y*SIZE*BLOCK_SIZE;

        for(unsigned int y = 0; y < SIZE; y++)
        {
            aabb.high_y = aabb.low_y + BLOCK_SIZE;

            aabb.low_z = this->z*SIZE*BLOCK_SIZE;

            for(unsigned int z = 0; z < SIZE; z++, aabb.low_z += BLOCK_SIZE)
            {
                aabb.high_z = aabb.low_z + BLOCK_SIZE;

                if(blocks[x][y][z] == BLOCK_AIR)
                    continue;

                if(aabb.intersects(other))
                    return true;
            }

            aabb.low_y += BLOCK_SIZE;
        }

        aabb.low_x += BLOCK_SIZE;
    }

    return false;
}

void Chunk::makeTree(unsigned int x, unsigned int y, unsigned int z)
{
    int max_height = SIZE - y;

    if(max_height < 5 || x < 1 || x > SIZE-2 || z < 1 || z > SIZE-2)
        return;

    max_height = std::min(max_height, (rand() & 0x5) + 4);

    max_height -= 3;

    for(unsigned int y1 = y; y1 < y + max_height; y1++)
        blocks[x][y1][z] = BLOCK_WOOD;

    for(unsigned int y1 = y + max_height; y1 < y + max_height + 2; y1++)
    {
        blocks[x][y1][z-1] = BLOCK_LEAVES;
        blocks[x-1][y1][z-1] = BLOCK_LEAVES;
        blocks[x+1][y1][z-1] = BLOCK_LEAVES;
        blocks[x][y1][z] = BLOCK_LEAVES;
        blocks[x-1][y1][z] = BLOCK_LEAVES;
        blocks[x+1][y1][z] = BLOCK_LEAVES;
        blocks[x][y1][z+1] = BLOCK_LEAVES;
        blocks[x-1][y1][z+1] = BLOCK_LEAVES;
        blocks[x+1][y1][z+1] = BLOCK_LEAVES;
    }

    blocks[x][y + max_height + 2][z] = BLOCK_LEAVES;
}
